from typing import Optional, Union

from dagster import (
    AssetsDefinition,
    AssetSpec,
    _check as check,
)
from dagster._core.definitions.utils import VALID_NAME_REGEX

MIGRATED_TAG = "airlift/task_migrated"
DAG_ID_TAG = "airlift/dag_id"
TASK_ID_TAG = "airlift/task_id"


def convert_to_valid_dagster_name(name: str) -> str:
    """Converts a name to a valid dagster name by replacing invalid characters with underscores. / is converted to a double underscore."""
    return "".join(c if VALID_NAME_REGEX.match(c) else "__" if c == "/" else "_" for c in name)


def get_task_id_from_asset(asset: Union[AssetsDefinition, AssetSpec]) -> Optional[str]:
    return _get_prop_from_asset(asset, TASK_ID_TAG, 1)


def get_dag_id_from_asset(asset: Union[AssetsDefinition, AssetSpec]) -> Optional[str]:
    return _get_prop_from_asset(asset, DAG_ID_TAG, 0)


def _get_prop_from_asset(
    asset: Union[AssetSpec, AssetsDefinition], prop_tag: str, position: int
) -> Optional[str]:
    prop_from_asset_tags = prop_from_tags(asset, prop_tag)
    if isinstance(asset, AssetSpec):
        return prop_from_asset_tags
    prop_from_op_tags = None
    if asset.node_def.tags and prop_tag in asset.node_def.tags:
        prop_from_op_tags = asset.node_def.tags[prop_tag]
    prop_from_name = None
    if len(asset.node_def.name.split("__")) == 2:
        prop_from_name = asset.node_def.name.split("__")[position]
    if prop_from_asset_tags and prop_from_op_tags:
        check.invariant(
            prop_from_asset_tags == prop_from_op_tags,
            f"ID mismatch between asset tags and op tags: {prop_from_asset_tags} != {prop_from_op_tags}",
        )
    if prop_from_asset_tags and prop_from_name:
        check.invariant(
            prop_from_asset_tags == prop_from_name,
            f"ID mismatch between tags and name: {prop_from_asset_tags} != {prop_from_name}",
        )
    if prop_from_op_tags and prop_from_name:
        check.invariant(
            prop_from_op_tags == prop_from_name,
            f"ID mismatch between op tags and name: {prop_from_op_tags} != {prop_from_name}",
        )
    return prop_from_asset_tags or prop_from_op_tags or prop_from_name


def prop_from_tags(asset: Union[AssetsDefinition, AssetSpec], prop_tag: str) -> Optional[str]:
    specs = asset.specs if isinstance(asset, AssetsDefinition) else [asset]
    asset_name = (
        asset.node_def.name if isinstance(asset, AssetsDefinition) else asset.key.to_user_string()
    )
    if any(prop_tag in spec.tags for spec in specs):
        prop = None
        for spec in specs:
            if prop is None:
                prop = spec.tags[prop_tag]
            else:
                if spec.tags.get(prop_tag) is None:
                    check.failed(f"Missing {prop_tag} tag in spec {spec.key} for {asset_name}")
                check.invariant(
                    prop == spec.tags[prop_tag],
                    f"Task ID mismatch within same AssetsDefinition: {prop} != {spec.tags[prop_tag]}",
                )
        return prop
    return None



from typing import Any, Optional, Tuple

import boto3
import requests

from ..core.airflow_instance import AirflowAuthBackend


def get_session_info(mwaa: Any, env_name: str) -> Tuple[str, str]:
    # Initialize MWAA client and request a web login token
    response = mwaa.create_web_login_token(Name=env_name)

    # Extract the web server hostname and login token
    web_server_host_name = response["WebServerHostname"]
    web_token = response["WebToken"]

    # Construct the URL needed for authentication
    login_url = f"https://{web_server_host_name}/aws_mwaa/login"
    login_payload = {"token": web_token}

    # Make a POST request to the MWAA login url using the login payload
    response = requests.post(login_url, data=login_payload, timeout=10)

    # Check if login was succesfull
    if response.status_code == 200:
        # Return the hostname and the session cookie
        return (web_server_host_name, response.cookies["session"])
    else:
        raise Exception(f"Failed to get session info: {response.text}")


class MwaaSessionAuthBackend(AirflowAuthBackend):
    def __init__(self, mwaa_client: Any, env_name: str) -> None:
        self.mwaa_client = mwaa_client
        self.env_name = env_name
        # Session info is generated when we either try to retrieve a session or retrieve the web server url
        self._session_info: Optional[Tuple[str, str]] = None

    @staticmethod
    def from_profile(region: str, env_name: str, profile_name: Optional[str] = None):
        boto_session = boto3.Session(profile_name=profile_name, region_name=region)
        mwaa = boto_session.client("mwaa")
        return MwaaSessionAuthBackend(mwaa_client=mwaa, env_name=env_name)

    def get_session(self) -> requests.Session:
        # Get the session info
        if not self._session_info:
            self._session_info = get_session_info(mwaa=self.mwaa_client, env_name=self.env_name)
        session_cookie = self._session_info[1]
        # Create a new session
        session = requests.Session()
        session.cookies.set("session", session_cookie)

        # Return the session
        return session

    def get_webserver_url(self) -> str:
        if not self._session_info:
            self._session_info = get_session_info(mwaa=self.mwaa_client, env_name=self.env_name)
        return f"https://{self._session_info[0]}"



from typing import Dict, Mapping, Union

from dagster import (
    AssetsDefinition,
    AssetSpec,
    Definitions,
    _check as check,
)

from dagster_airlift.core.utils import DAG_ID_TAG, TASK_ID_TAG


class TaskDefs:
    def __init__(self, task_id: str, defs: Definitions):
        self.task_id = task_id
        self.defs = defs


def apply_tags_to_all_specs(defs: Definitions, tags: Dict[str, str]) -> Definitions:
    return Definitions(
        assets=[
            # Right now we make assumptions that we only support AssetSpec and AssetsDefinition
            # in orchesrated_defs.
            # https://linear.app/dagster-labs/issue/FOU-369/support-cacheableassetsdefinition-and-sourceasset-in-airlift
            assets_def_with_af_tags(
                check.inst(
                    asset,
                    (AssetSpec, AssetsDefinition),
                    "Only supports AssetSpec and AssetsDefinition right now",
                ),
                tags,
            )
            for asset in (defs.assets or [])
        ],
        resources=defs.resources,
        sensors=defs.sensors,
        schedules=defs.schedules,
        jobs=defs.jobs,
        loggers=defs.loggers,
        executor=defs.executor,
        asset_checks=defs.asset_checks,
    )


def spec_with_tags(spec: AssetSpec, tags: Mapping[str, str]) -> "AssetSpec":
    return spec._replace(tags={**spec.tags, **tags})


def assets_def_with_af_tags(
    assets_def: Union[AssetsDefinition, AssetSpec], tags: Mapping[str, str]
) -> Union[AssetsDefinition, AssetSpec]:
    return (
        assets_def.map_asset_specs(lambda spec: spec_with_tags(spec, tags))
        if isinstance(assets_def, AssetsDefinition)
        else spec_with_tags(assets_def, tags)
    )


def dag_defs(dag_id: str, *defs: TaskDefs) -> Definitions:
    """Construct a Dagster :py:class:`Definitions` object with definitions
    associated with a particular Dag in Airflow that is being tracked by Airlift tooling.

    Concretely this adds tags to all asset specs in the provided definitions
    with the provided dag_id and task_id. Dag id is tagged with the
    "airlift/dag_id" key and task id is tagged with the "airlift/task_id" key.

    Used in concert with :py:func:`task_defs`.

    Example:
    .. code-block:: python
        defs = dag_defs(
            "dag_one",
            task_defs("task_one", Definitions(assets=[AssetSpec(key="asset_one"]))),
            task_defs("task_two", Definitions(assets=[AssetSpec(key="asset_two"), AssetSpec(key="asset_three")])),
        )
    """
    defs_to_merge = []
    for task_def in defs:
        defs_to_merge.append(
            apply_tags_to_all_specs(
                defs=task_def.defs,
                tags={DAG_ID_TAG: dag_id, TASK_ID_TAG: task_def.task_id},
            )
        )
    return Definitions.merge(*defs_to_merge)


def task_defs(task_id, defs: Definitions) -> TaskDefs:
    """Associate a set of definitions with a particular task in Airflow that is being tracked
    by Airlift tooling.
    """
    return TaskDefs(task_id, defs)



import inspect
import logging
import os
from typing import Any, Callable, Dict, Set, Tuple

import requests
from airflow.models.operator import BaseOperator
from airflow.operators.python import PythonOperator

from dagster_airlift.core.utils import DAG_ID_TAG, TASK_ID_TAG

from .gql_queries import ASSET_NODES_QUERY, RUNS_QUERY, TRIGGER_ASSETS_MUTATION

logger = logging.getLogger(__name__)


def compute_fn() -> None:
    # https://github.com/apache/airflow/discussions/24463
    os.environ["NO_PROXY"] = "*"
    dag_id = os.environ["AIRFLOW_CTX_DAG_ID"]
    task_id = os.environ["AIRFLOW_CTX_TASK_ID"]
    dagster_url = os.environ["DAGSTER_URL"]
    return launch_runs_for_task(dag_id, task_id, dagster_url)


def launch_runs_for_task(dag_id: str, task_id: str, dagster_url: str) -> None:
    expected_op_name = f"{dag_id}__{task_id}"

    assets_to_trigger = {}  # key is (repo_location, repo_name, job_name), value is list of asset keys
    # create graphql client
    response = requests.post(f"{dagster_url}/graphql", json={"query": ASSET_NODES_QUERY}, timeout=3)
    for asset_node in response.json()["data"]["assetNodes"]:
        tags = {tag["key"]: tag["value"] for tag in asset_node["tags"]}
        # match assets based on conventional dag_id__task_id naming or based on explicit tags
        if asset_node["opName"] == expected_op_name or (
            tags.get(DAG_ID_TAG) == dag_id and tags.get(TASK_ID_TAG) == task_id
        ):
            repo_location = asset_node["jobs"][0]["repository"]["location"]["name"]
            repo_name = asset_node["jobs"][0]["repository"]["name"]
            job_name = asset_node["jobs"][0]["name"]
            if (repo_location, repo_name, job_name) not in assets_to_trigger:
                assets_to_trigger[(repo_location, repo_name, job_name)] = []
            assets_to_trigger[(repo_location, repo_name, job_name)].append(
                asset_node["assetKey"]["path"]
            )
    logger.debug(f"Found assets to trigger: {assets_to_trigger}")
    triggered_runs = []
    for (repo_location, repo_name, job_name), asset_keys in assets_to_trigger.items():
        execution_params = {
            "mode": "default",
            "executionMetadata": {"tags": []},
            "runConfigData": "{}",
            "selector": {
                "repositoryLocationName": repo_location,
                "repositoryName": repo_name,
                "pipelineName": job_name,
                "assetSelection": [{"path": asset_key} for asset_key in asset_keys],
                "assetCheckSelection": [],
            },
        }
        logger.debug(
            f"Triggering run for {repo_location}/{repo_name}/{job_name} with assets {asset_keys}"
        )
        response = requests.post(
            f"{dagster_url}/graphql",
            json={
                "query": TRIGGER_ASSETS_MUTATION,
                "variables": {"executionParams": execution_params},
            },
            timeout=3,
        )
        run_id = response.json()["data"]["launchPipelineExecution"]["run"]["id"]
        logger.debug(f"Launched run {run_id}...")
        triggered_runs.append(run_id)
    completed_runs = {}  # key is run_id, value is status
    while len(completed_runs) < len(triggered_runs):
        for run_id in triggered_runs:
            if run_id in completed_runs:
                continue
            response = requests.post(
                f"{dagster_url}/graphql",
                json={"query": RUNS_QUERY, "variables": {"runId": run_id}},
                timeout=3,
            )
            run_status = response.json()["data"]["runOrError"]["status"]
            if run_status in ["SUCCESS", "FAILURE", "CANCELED"]:
                logger.debug(f"Run {run_id} completed with status {run_status}")
                completed_runs[run_id] = run_status
    non_successful_runs = [
        run_id for run_id, status in completed_runs.items() if status != "SUCCESS"
    ]
    if non_successful_runs:
        raise Exception(f"Runs {non_successful_runs} did not complete successfully.")
    logger.debug("All runs completed successfully.")
    return None


class DagsterOperator(PythonOperator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, python_callable=compute_fn)


def build_dagster_task(original_task: BaseOperator) -> DagsterOperator:
    return instantiate_dagster_operator(original_task)


def instantiate_dagster_operator(original_task: BaseOperator) -> DagsterOperator:
    """Instantiates a DagsterOperator as a copy of the provided airflow task.

    We attempt to copy as many of the original task's attributes as possible, while respecting
    that attributes may change between airflow versions. In order to do this, we inspect the
    arguments available to the BaseOperator constructor and copy over any of those arguments that
    are available as attributes on the original task.
    This approach has limitations:
    - If the task attribute is transformed and stored on another property, it will not be copied.
    - If the task attribute is transformed in a way that makes it incompatible with the constructor arg
    and stored in the same property, that will attempt to be copied and potentiall break initialization.
    In the future, if we hit problems with this, we may need to add argument overrides to ensure we either
    attempt to include certain additional attributes, or exclude others. If this continues to be a problem
    across airflow versions, it may be necessary to revise this approach to one that explicitly maps airflow
    version to a set of expected arguments and attributes.
    """
    base_operator_args, base_operator_args_with_defaults = get_params(BaseOperator.__init__)
    init_kwargs = {}

    ignore_args = ["kwargs", "args", "dag"]
    for arg in base_operator_args:
        if arg in ignore_args or getattr(original_task, arg, None) is None:
            continue
        init_kwargs[arg] = getattr(original_task, arg)
    for kwarg, default in base_operator_args_with_defaults.items():
        if kwarg in ignore_args or getattr(original_task, kwarg, None) is None:
            continue
        init_kwargs[kwarg] = getattr(original_task, kwarg, default)

    return DagsterOperator(**init_kwargs)


def get_params(func: Callable[..., Any]) -> Tuple[Set[str], Dict[str, Any]]:
    """Retrieves the args and kwargs from the signature of a given function or method.
    For kwargs, default values are retrieved as well.

    Args:
        func (Callable[..., Any]): The function or method to inspect.

    Returns:
        Tuple[Set[str], Dict[str, Any]]:
            - A set of argument names that do not have default values.
            - A dictionary of keyword argument names and their default values.
    """
    # Get the function's signature
    sig = inspect.signature(func)

    # Initialize sets for args without defaults and kwargs with defaults
    args_with_defaults = {}
    args = set()

    # Iterate over function parameters
    for name, param in sig.parameters.items():
        if param.default is inspect.Parameter.empty and name != "self":  # Exclude 'self'
            args.add(name)
        else:
            if name != "self":  # Exclude 'self'
                args_with_defaults[name] = param.default

    return args, args_with_defaults



import requests

from .airflow_instance import AirflowAuthBackend


class BasicAuthBackend(AirflowAuthBackend):
    def __init__(self, webserver_url: str, username: str, password: str):
        self._webserver_url = webserver_url
        self.username = username
        self.password = password

    def get_session(self) -> requests.Session:
        session = requests.Session()
        session.auth = (self.username, self.password)
        return session

    def get_webserver_url(self) -> str:
        return self._webserver_url



from collections import defaultdict
from datetime import timedelta
from typing import Dict, List, Sequence, Set, Tuple

from dagster import (
    AssetCheckKey,
    AssetKey,
    AssetMaterialization,
    DefaultSensorStatus,
    JsonMetadataValue,
    MarkdownMetadataValue,
    RunRequest,
    SensorDefinition,
    SensorEvaluationContext,
    SensorResult,
    TimestampMetadataValue,
    _check as check,
    sensor,
)
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.repository_definition.repository_definition import (
    RepositoryDefinition,
)
from dagster._core.utils import toposort_flatten
from dagster._record import record
from dagster._time import datetime_from_timestamp, get_current_datetime, get_current_timestamp

from .airflow_instance import AirflowInstance, TaskInstance
from .utils import MIGRATED_TAG, get_dag_id_from_asset, get_task_id_from_asset


def build_airflow_polling_sensor(
    airflow_instance: AirflowInstance,
) -> SensorDefinition:
    @sensor(
        name="airflow_dag_status_sensor",
        minimum_interval_seconds=1,
        default_status=DefaultSensorStatus.RUNNING,
        # Tracking bug around this here: https://linear.app/dagster-labs/issue/FOU-376/make-airflow-polling-sensor-work-with-asset-selection=
        asset_selection=AssetSelection.assets(),
    )
    def airflow_dag_sensor(context: SensorEvaluationContext) -> SensorResult:
        """Sensor to report materialization events for each asset as new runs come in."""
        repository_def = check.not_none(context.repository_def)
        last_effective_date = (
            datetime_from_timestamp(float(context.cursor))
            if context.cursor
            else get_current_datetime() - timedelta(days=1)
        )
        current_date = get_current_datetime()
        materializations_to_report: List[Tuple[float, AssetMaterialization]] = []
        toposorted_keys = toposorted_asset_keys(repository_def)
        asset_check_keys_to_request = set()
        unmigrated_info = get_unmigrated_info(repository_def)
        for dag_id, peered_dag_asset_info in unmigrated_info.asset_info_by_dag_id.items():
            dag_key = peered_dag_asset_info.dag_asset_key
            task_keys = peered_dag_asset_info.task_asset_keys
            # For now, we materialize assets representing tasks only when the whole dag completes.
            # With a more robust cursor that can let us know when we've seen a particular task run already, then we can relax this constraint.
            for dag_run in airflow_instance.get_dag_runs(dag_id, last_effective_date, current_date):
                if not dag_run.success:
                    raise Exception("Should only see successful dag runs at this point.")

                metadata = {
                    "Airflow Run ID": dag_run.run_id,
                    "Run Metadata (raw)": JsonMetadataValue(dag_run.metadata),
                    "Run Type": dag_run.run_type,
                    "Airflow Config": JsonMetadataValue(dag_run.config),
                    "Creation Timestamp": TimestampMetadataValue(get_current_timestamp()),
                }
                # Add dag materialization
                dag_metadata = {
                    **metadata,
                    "Run Details": MarkdownMetadataValue(f"[View Run]({dag_run.url})"),
                    "Start Date": TimestampMetadataValue(dag_run.start_date),
                    "End Date": TimestampMetadataValue(dag_run.end_date),
                }
                materializations_to_report.append(
                    (
                        dag_run.end_date,
                        AssetMaterialization(
                            asset_key=dag_key,
                            description=dag_run.note,
                            metadata=dag_metadata,
                        ),
                    )
                )
                asset_check_keys_to_request.update(unmigrated_info.checks_per_key[dag_key])
                task_runs = {}
                for task_id, asset_key in task_keys:
                    task_run: TaskInstance = task_runs.get(
                        task_id, airflow_instance.get_task_instance(dag_id, task_id, dag_run.run_id)
                    )
                    task_runs[task_id] = task_run
                    task_metadata = {
                        **metadata,
                        "Run Details": MarkdownMetadataValue(f"[View Run]({task_run.details_url})"),
                        "Task Logs": MarkdownMetadataValue(f"[View Logs]({task_run.log_url})"),
                        "Start Date": TimestampMetadataValue(task_run.start_date),
                        "End Date": TimestampMetadataValue(task_run.end_date),
                    }
                    materializations_to_report.append(
                        (
                            task_run.end_date,
                            AssetMaterialization(
                                asset_key=asset_key,
                                description=task_run.note,
                                metadata=task_metadata,
                            ),
                        )
                    )
                    asset_check_keys_to_request.update(unmigrated_info.checks_per_key[asset_key])
        # Sort materializations by end date and toposort order
        sorted_mats = sorted(
            materializations_to_report, key=lambda x: (x[0], toposorted_keys.index(x[1].asset_key))
        )
        context.update_cursor(str(current_date.timestamp()))
        return SensorResult(
            asset_events=[sorted_mat[1] for sorted_mat in sorted_mats],
            run_requests=[
                RunRequest(asset_check_keys=list(asset_check_keys_to_request), asset_selection=[])
            ]
            if asset_check_keys_to_request
            else None,
        )

    return airflow_dag_sensor


@record
class PeeredDagAssetInfo:
    dag_asset_key: AssetKey
    task_asset_keys: Set[Tuple[str, AssetKey]]


@record
class UnmigratedInfo:
    asset_info_by_dag_id: Dict[str, PeeredDagAssetInfo]
    checks_per_key: Dict[AssetKey, Set[AssetCheckKey]]


def get_unmigrated_info(
    repository_def: RepositoryDefinition,
) -> UnmigratedInfo:
    """For each dag, retrieve the list of asset keys which correspond, and are unmigrated.
    The key representing the "peered" dag will always be retrieved, but assets whose tasks are marked as "migrated" will not.
    """
    # First, we need to retrieve the upstreams for each asset key
    key_per_dag: Dict[str, AssetKey] = {}
    task_keys_per_dag = defaultdict(set)
    checks_per_key = defaultdict(set)

    for assets_def in repository_def.assets_defs_by_key.values():
        # We could be more specific about the checks here to ensure that there's only one asset key
        # specifying the dag, and that all others have a task id.
        dag_id = get_dag_id_from_asset(assets_def)
        task_id = get_task_id_from_asset(assets_def)
        if dag_id is None:
            continue
        if task_id is None:
            key_per_dag[dag_id] = (
                assets_def.key
            )  # There should only be one key in the case of a "dag" asset
        else:
            migrated = (
                MIGRATED_TAG in assets_def.node_def.tags
                and assets_def.node_def.tags[MIGRATED_TAG] == "True"
            )
            if migrated:
                continue
            else:
                task_keys_per_dag[dag_id].update((task_id, spec.key) for spec in assets_def.specs)
    for asset_check_key in repository_def.asset_checks_defs_by_key.keys():
        checks_per_key[asset_check_key.asset_key].add(asset_check_key)

    per_dag_asset_info = {
        dag_id: PeeredDagAssetInfo(dag_asset_key=key, task_asset_keys=task_keys_per_dag[dag_id])
        for dag_id, key in key_per_dag.items()
    }
    return UnmigratedInfo(asset_info_by_dag_id=per_dag_asset_info, checks_per_key=checks_per_key)


def toposorted_asset_keys(
    repository_def: RepositoryDefinition,
) -> Sequence[AssetKey]:
    asset_dep_graph = defaultdict(set)  # upstreams
    for assets_def in repository_def.assets_defs_by_key.values():
        for spec in assets_def.specs:
            asset_dep_graph[spec.key].update(dep.asset_key for dep in spec.deps)

    return toposort_flatten(asset_dep_graph)



from ..migration_state import load_migration_state_from_yaml as load_migration_state_from_yaml
from .basic_auth import BasicAuthBackend as BasicAuthBackend
from .dag_defs import (
    dag_defs as dag_defs,
    task_defs as task_defs,
)
from .defs_builders import specs_from_task as specs_from_task
from .defs_from_airflow import (
    AirflowInstance as AirflowInstance,
    build_defs_from_airflow_instance as build_defs_from_airflow_instance,
)



import json
import logging
from typing import Any, Dict, List, Optional

from airflow import DAG
from airflow.models import BaseOperator

from dagster_airlift.in_airflow.dagster_operator import build_dagster_task

from ..migration_state import AirflowMigrationState


def mark_as_dagster_migrating(
    *,
    global_vars: Dict[str, Any],
    migration_state: AirflowMigrationState,
    logger: Optional[logging.Logger] = None,
) -> None:
    """Alters all airflow dags in the current context to be marked as migrating to dagster.
    Uses a migration dictionary to determine the status of the migration for each task within each dag.
    Should only ever be the last line in a dag file.

    Args:
        global_vars (Dict[str, Any]): The global variables in the current context. In most cases, retrieved with `globals()` (no import required).
            This is equivalent to what airflow already does to introspect the dags which exist in a given module context:
            https://airflow.apache.org/docs/apache-airflow/stable/core-concepts/dags.html#loading-dags
        migration_state (AirflowMigrationState): The migration state for the dags.
        logger (Optional[logging.Logger]): The logger to use. Defaults to logging.getLogger("dagster_airlift").
    """
    caller_module = global_vars.get("__module__")
    suffix = f" in module `{caller_module}`" if caller_module else ""
    if not logger:
        logger = logging.getLogger("dagster_airlift")
    logger.debug(f"Searching for dags migrating to dagster{suffix}...")
    migrating_dags: List[DAG] = []
    # Do a pass to collect dags and ensure that migration information is set correctly.
    for obj in global_vars.values():
        if not isinstance(obj, DAG):
            continue
        dag: DAG = obj
        if not migration_state.dag_has_migration_state(dag.dag_id):
            logger.debug(f"Dag with id `{dag.dag_id}` has no migration state. Skipping...")
            continue
        logger.debug(f"Dag with id `{dag.dag_id}` has migration state.")
        migration_state_for_dag = migration_state.dags[dag.dag_id]
        for task_id in migration_state_for_dag.tasks.keys():
            if task_id not in dag.task_dict:
                raise Exception(
                    f"Task with id `{task_id}` not found in dag `{dag.dag_id}`. Found tasks: {list(dag.task_dict.keys())}"
                )
            if not isinstance(dag.task_dict[task_id], BaseOperator):
                raise Exception(
                    f"Task with id `{task_id}` in dag `{dag.dag_id}` is not an instance of BaseOperator. This likely means a MappedOperator was attempted, which is not yet supported by airlift."
                )
        migrating_dags.append(dag)

    for dag in migrating_dags:
        logger.debug(f"Tagging dag {dag.dag_id} as migrating.")
        dag.tags.append(
            json.dumps(
                {"DAGSTER_MIGRATION_STATUS": migration_state.get_migration_dict_for_dag(dag.dag_id)}
            )
        )
        migration_state_for_dag = migration_state.dags[dag.dag_id]
        migrated_tasks = set()
        for task_id, task_state in migration_state_for_dag.tasks.items():
            if not task_state.migrated:
                logger.debug(
                    f"Task {task_id} in dag {dag.dag_id} has `migrated` set to False. Skipping..."
                )
                continue

            # At this point, we should be assured that the task exists within the task_dict of the dag, and is a BaseOperator.
            original_op: BaseOperator = dag.task_dict[task_id]  # type: ignore  # we already confirmed this is BaseOperator
            del dag.task_dict[task_id]
            if original_op.task_group is not None:
                del original_op.task_group.children[task_id]
            logger.debug(
                f"Creating new operator for task {original_op.task_id} in dag {original_op.dag_id}"
            )
            new_op = build_dagster_task(original_op)
            original_op.dag.task_dict[original_op.task_id] = new_op

            new_op.upstream_task_ids = original_op.upstream_task_ids
            new_op.downstream_task_ids = original_op.downstream_task_ids
            new_op.dag = original_op.dag
            original_op.dag = None
            migrated_tasks.add(task_id)
        logger.debug(f"Migrated tasks {migrated_tasks} in dag {dag.dag_id}.")
    logging.debug(f"Migrated {len(migrating_dags)}.")
    logging.debug(f"Completed marking dags and tasks as migrating to dagster{suffix}.")



from typing import Sequence, Union

from dagster import AssetSpec
from dagster._core.definitions.asset_key import CoercibleToAssetKey
from typing_extensions import TypeAlias

from .utils import DAG_ID_TAG, TASK_ID_TAG

CoercibleToAssetSpec: TypeAlias = Union[AssetSpec, CoercibleToAssetKey]


def specs_from_task(
    *, task_id: str, dag_id: str, assets: Sequence[CoercibleToAssetSpec]
) -> Sequence[AssetSpec]:
    """Construct a Dagster :py:class:`Definitions` object from a provided set of assets,
    with a mapping to which airflow task produces those assets.
    """
    return [
        asset
        if isinstance(asset, AssetSpec)
        else AssetSpec(key=asset, tags={DAG_ID_TAG: dag_id, TASK_ID_TAG: task_id})
        for asset in assets
    ]



from typing import Optional

import yaml
from dagster import AssetExecutionContext, Definitions
from dagster_dbt import (
    DagsterDbtTranslator,
    DagsterDbtTranslatorSettings,
    DbtCliResource,
    DbtProject,
    dbt_assets,
)
from dagster_dbt.dbt_manifest import DbtManifestParam


def dbt_defs(
    *,
    manifest: DbtManifestParam,
    project: DbtProject,
    name: Optional[str] = None,
) -> Definitions:
    if not name:
        with open(project.project_dir.joinpath("dbt_project.yml")) as file:
            dbt_project_yml = yaml.safe_load(file)
            if "name" not in dbt_project_yml:
                raise ValueError("name not found in dbt_project.yml")
            if not dbt_project_yml["name"]:
                raise ValueError("name in dbt_project.yml is empty")
            name = f"build_{dbt_project_yml['name']}"

    @dbt_assets(
        name=name,
        manifest=manifest,
        project=project,
        dagster_dbt_translator=DagsterDbtTranslator(
            settings=DagsterDbtTranslatorSettings(enable_asset_checks=False)
        ),
    )
    def _dbt_asset(context: AssetExecutionContext, dbt: DbtCliResource):
        context.log.info(f"project_dir {dbt.project_dir}")
        yield from dbt.cli(["build"], context=context).stream()

    return Definitions(
        assets=[_dbt_asset],
        resources={"dbt": DbtCliResource(project_dir=project)},
    )



from typing import Optional

from dagster import Definitions

from dagster_airlift.core.sensor import build_airflow_polling_sensor
from dagster_airlift.migration_state import AirflowMigrationState

from .airflow_cacheable_assets_def import DEFAULT_POLL_INTERVAL, AirflowCacheableAssetsDefinition
from .airflow_instance import AirflowInstance


def build_defs_from_airflow_instance(
    airflow_instance: AirflowInstance,
    cache_polling_interval: int = DEFAULT_POLL_INTERVAL,
    defs: Optional[Definitions] = None,
    # This parameter will go away once we can derive the migration state from airflow itself, using our built in utilities.
    # Alternatively, we can keep it around to let people override the migration state if they want.
    migration_state_override: Optional[AirflowMigrationState] = None,
) -> Definitions:
    """From a provided airflow instance and a set of airflow-orchestrated dagster definitions, build a set of dagster definitions to peer and observe the airflow instance.

    Each airflow dag will be represented as a dagster asset, and each dag run will be represented as an asset materialization. A :py:class:`dagster.SensorDefinition` provided in the returned Definitions will be used to poll for dag runs.

    The provided orchestrated defs are expected to contain fully qualified :py:class:`dagster.AssetsDefinition` objects, each of which should be mapped to a task and dag in the provided airflow instance. Using the airflow instance,
    dagster will provide dependency information between the assets representing tasks, and the dags that they contain. The included :py:class:`dagster.SensorDefinition` will poll for dag runs and materialize runs including each task as an asset for that task.

    There are two ways that the mapping can be done on a provided definition:
    1. By using the `airlift/dag_id` and `airlift/task_id` op tags on the underlying :py:class:`dagster.NodeDefinition` for the asset.
    2. By using an opinionated naming format on the :py:class:`dagster.NodeDefinition` for the asset. The naming format is `dag_id__task_id`.

    Args:
        airflow_instance (AirflowInstance): The airflow instance to peer with.
        cache_polling_interval (int, optional): The interval at which to poll for updated assets from airflow. Defaults to 60 seconds.
        defs (Optional[Definitions], optional): Additional definitions to pass to the overall created Definitions object. Assets provided here can be mapped to specific tasks and dags within the provided airflow instance. Defaults to None.
        migration_state_override (Optional[AirflowMigrationState], optional): The migration state to use for the provided orchestrated defs. Defaults to None.

    Returns:
        Definitions: The definitions to use for the provided airflow instance. Contains an asset per dag, an asset per task in the provided orchestrated defs, all resources provided in the orchestrated defs, and a sensor to poll for airflow dag runs.

    """
    assets_defs = AirflowCacheableAssetsDefinition(
        airflow_instance=airflow_instance,
        defs=defs,
        poll_interval=cache_polling_interval,
        migration_state_override=migration_state_override,
    )
    # Now, we construct the sensor that will poll airflow for dag runs.
    airflow_sensor = build_airflow_polling_sensor(airflow_instance=airflow_instance)
    return Definitions(
        assets=[assets_defs],
        asset_checks=defs.asset_checks if defs else None,
        sensors=[airflow_sensor, *defs.sensors] if defs and defs.sensors else [airflow_sensor],
        schedules=defs.schedules if defs else None,
        jobs=defs.jobs if defs else None,
        executor=defs.executor if defs else None,
        loggers=defs.loggers if defs else None,
        resources=defs.resources if defs else None,
    )



from typing import Callable, Optional, Sequence

from dagster import AssetSpec, Definitions
from dagster._core.definitions.decorators.asset_decorator import multi_asset


def defs_for_python_callable(
    *, asset_specs: Sequence[AssetSpec], python_callable: Callable, name: Optional[str] = None
) -> Definitions:
    name = name or python_callable.__name__

    @multi_asset(name=name, specs=asset_specs)
    def _an_asset() -> None:
        python_callable()

    return Definitions(assets=[_an_asset])



import datetime
import json
from abc import ABC
from functools import cached_property
from typing import Any, Dict, List

import requests
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.errors import DagsterError
from dagster._record import record
from dagster._time import get_current_datetime

from ..migration_state import DagMigrationState
from .utils import convert_to_valid_dagster_name

TERMINAL_STATES = {"success", "failed", "skipped", "up_for_retry", "up_for_reschedule"}


class AirflowAuthBackend(ABC):
    def get_session(self) -> requests.Session:
        raise NotImplementedError("This method must be implemented by subclasses.")

    def get_webserver_url(self) -> str:
        raise NotImplementedError("This method must be implemented by subclasses.")


class AirflowInstance:
    def __init__(self, auth_backend: AirflowAuthBackend, name: str) -> None:
        self.auth_backend = auth_backend
        self.name = name

    @property
    def normalized_name(self) -> str:
        return self.name.replace(" ", "_").replace("-", "_")

    def get_api_url(self) -> str:
        return f"{self.auth_backend.get_webserver_url()}/api/v1"

    def list_dags(self) -> List["DagInfo"]:
        response = self.auth_backend.get_session().get(f"{self.get_api_url()}/dags")
        if response.status_code == 200:
            dags = response.json()
            webserver_url = self.auth_backend.get_webserver_url()
            return [
                DagInfo(
                    webserver_url=webserver_url,
                    dag_id=dag["dag_id"],
                    metadata=dag,
                )
                for dag in dags["dags"]
            ]
        else:
            raise DagsterError(
                f"Failed to fetch DAGs. Status code: {response.status_code}, Message: {response.text}"
            )

    def get_task_instance(self, dag_id: str, task_id: str, run_id: str) -> "TaskInstance":
        response = self.auth_backend.get_session().get(
            f"{self.get_api_url()}/dags/{dag_id}/dagRuns/{run_id}/taskInstances/{task_id}"
        )
        if response.status_code == 200:
            return TaskInstance(
                webserver_url=self.auth_backend.get_webserver_url(),
                dag_id=dag_id,
                task_id=task_id,
                run_id=run_id,
                metadata=response.json(),
            )
        else:
            raise DagsterError(
                f"Failed to fetch task instance for {dag_id}/{task_id}/{run_id}. Status code: {response.status_code}, Message: {response.text}"
            )

    def get_task_info(self, dag_id: str, task_id: str) -> "TaskInfo":
        response = self.auth_backend.get_session().get(
            f"{self.get_api_url()}/dags/{dag_id}/tasks/{task_id}"
        )
        if response.status_code == 200:
            return TaskInfo(
                webserver_url=self.auth_backend.get_webserver_url(),
                dag_id=dag_id,
                task_id=task_id,
                metadata=response.json(),
            )
        else:
            raise DagsterError(
                f"Failed to fetch task info for {dag_id}/{task_id}. Status code: {response.status_code}, Message: {response.text}"
            )

    def get_dag_source_code(self, file_token: str) -> str:
        response = self.auth_backend.get_session().get(
            f"{self.get_api_url()}/dagSources/{file_token}"
        )
        if response.status_code == 200:
            return response.text
        else:
            raise DagsterError(
                f"Failed to fetch source code. Status code: {response.status_code}, Message: {response.text}"
            )

    @staticmethod
    def airflow_str_from_datetime(dt: datetime.datetime) -> str:
        return dt.strftime("%Y-%m-%dT%H:%M:%S+00:00")

    def get_dag_runs(
        self, dag_id: str, start_date: datetime.datetime, end_date: datetime.datetime
    ) -> List["DagRun"]:
        response = self.auth_backend.get_session().get(
            f"{self.get_api_url()}/dags/{dag_id}/dagRuns",
            params={
                "updated_at_gte": self.airflow_str_from_datetime(start_date),
                "updated_at_lte": self.airflow_str_from_datetime(end_date),
                "state": ["success"],
            },
        )
        if response.status_code == 200:
            webserver_url = self.auth_backend.get_webserver_url()
            return [
                DagRun(
                    webserver_url=webserver_url,
                    dag_id=dag_id,
                    run_id=dag_run["dag_run_id"],
                    metadata=dag_run,
                )
                for dag_run in response.json()["dag_runs"]
            ]
        else:
            raise DagsterError(
                f"Failed to fetch dag runs for {dag_id}. Status code: {response.status_code}, Message: {response.text}"
            )

    def trigger_dag(self, dag_id: str) -> str:
        response = self.auth_backend.get_session().post(
            f"{self.get_api_url()}/dags/{dag_id}/dagRuns",
            json={},
        )
        if response.status_code != 200:
            raise DagsterError(
                f"Failed to launch run for {dag_id}. Status code: {response.status_code}, Message: {response.text}"
            )
        return response.json()["dag_run_id"]

    def get_dag_run(self, dag_id: str, run_id: str) -> "DagRun":
        response = self.auth_backend.get_session().get(
            f"{self.get_api_url()}/dags/{dag_id}/dagRuns/{run_id}"
        )
        if response.status_code != 200:
            raise DagsterError(
                f"Failed to fetch dag run for {dag_id}/{run_id}. Status code: {response.status_code}, Message: {response.text}"
            )
        return DagRun(
            webserver_url=self.auth_backend.get_webserver_url(),
            dag_id=dag_id,
            run_id=run_id,
            metadata=response.json(),
        )

    def wait_for_run_completion(self, dag_id: str, run_id: str, timeout: int = 30) -> None:
        start_time = get_current_datetime()
        while get_current_datetime() - start_time < datetime.timedelta(seconds=timeout):
            dag_run = self.get_dag_run(dag_id, run_id)
            if dag_run.finished:
                return
        raise DagsterError(f"Timed out waiting for airflow run {run_id} to finish.")

    @staticmethod
    def timestamp_from_airflow_date(airflow_date: str) -> float:
        try:
            return datetime.datetime.strptime(airflow_date, "%Y-%m-%dT%H:%M:%S+00:00").timestamp()
        except ValueError:
            return datetime.datetime.strptime(
                airflow_date, "%Y-%m-%dT%H:%M:%S.%f+00:00"
            ).timestamp()


@record
class DagInfo:
    webserver_url: str
    dag_id: str
    metadata: Dict[str, Any]

    @property
    def url(self) -> str:
        return f"{self.webserver_url}/dags/{self.dag_id}"

    @cached_property
    def dagster_safe_dag_id(self) -> str:
        """Name based on the dag_id that is safe to use in dagster."""
        return convert_to_valid_dagster_name(self.dag_id)

    @property
    def dag_asset_key(self) -> AssetKey:
        # Conventional asset key representing a successful run of an airfow dag.
        return AssetKey(["airflow_instance", "dag", self.dagster_safe_dag_id])

    @property
    def migration_state(self) -> DagMigrationState:
        tags = self.metadata.get("tags") or []
        migration_tag = next(
            (tag for tag in tags if "DAGSTER_MIGRATION_STATUS" in tag["name"]), None
        )
        if migration_tag:
            migration_dict = json.loads(migration_tag["name"])["DAGSTER_MIGRATION_STATUS"]
            return DagMigrationState.from_dict(migration_dict)
        return DagMigrationState(tasks={})


@record
class TaskInfo:
    webserver_url: str
    dag_id: str
    task_id: str
    metadata: Dict[str, Any]

    @property
    def dag_url(self) -> str:
        return f"{self.webserver_url}/dags/{self.dag_id}"


@record
class TaskInstance:
    webserver_url: str
    dag_id: str
    task_id: str
    run_id: str
    metadata: Dict[str, Any]

    @property
    def note(self) -> str:
        return self.metadata.get("note") or ""

    @property
    def details_url(self) -> str:
        return f"{self.webserver_url}/dags/{self.dag_id}/grid?dag_run_id={self.run_id}&task_id={self.task_id}"

    @property
    def log_url(self) -> str:
        return f"{self.details_url}&tab=logs"

    @property
    def start_date(self) -> float:
        return AirflowInstance.timestamp_from_airflow_date(self.metadata["start_date"])

    @property
    def end_date(self) -> float:
        return AirflowInstance.timestamp_from_airflow_date(self.metadata["end_date"])


@record
class DagRun:
    webserver_url: str
    dag_id: str
    run_id: str
    metadata: Dict[str, Any]

    @property
    def note(self) -> str:
        return self.metadata.get("note") or ""

    @property
    def url(self) -> str:
        return f"{self.webserver_url}/dags/{self.dag_id}/grid?dag_run_id={self.run_id}&tab=details"

    @property
    def success(self) -> bool:
        return self.metadata["state"] == "success"

    @property
    def finished(self) -> bool:
        return self.metadata["state"] in TERMINAL_STATES

    @property
    def run_type(self) -> str:
        return self.metadata["run_type"]

    @property
    def config(self) -> Dict[str, Any]:
        return self.metadata["conf"]

    @property
    def start_date(self) -> float:
        return AirflowInstance.timestamp_from_airflow_date(self.metadata["start_date"])

    @property
    def end_date(self) -> float:
        return AirflowInstance.timestamp_from_airflow_date(self.metadata["end_date"])



import hashlib
from collections import defaultdict
from typing import Any, Dict, List, Mapping, Optional, Sequence, Set, Union

from dagster import (
    AssetDep,
    AssetKey,
    AssetsDefinition,
    AssetSpec,
    Definitions,
    JsonMetadataValue,
    MarkdownMetadataValue,
    _check as check,
    multi_asset,
)
from dagster._core.definitions.cacheable_assets import (
    AssetsDefinitionCacheableData,
    CacheableAssetsDefinition,
)
from dagster._core.definitions.metadata.metadata_value import UrlMetadataValue
from dagster._record import record
from dagster._serdes import deserialize_value, serialize_value, whitelist_for_serdes
from dagster._serdes.serdes import (
    FieldSerializer,
    JsonSerializableValue,
    PackableValue,
    SerializableNonScalarKeyMapping,
    UnpackContext,
    WhitelistMap,
    pack_value,
    unpack_value,
)

from dagster_airlift.core.utils import convert_to_valid_dagster_name
from dagster_airlift.migration_state import AirflowMigrationState

from .airflow_instance import AirflowInstance, DagInfo, TaskInfo
from .utils import (
    DAG_ID_TAG,
    MIGRATED_TAG,
    TASK_ID_TAG,
    get_dag_id_from_asset,
    get_task_id_from_asset,
)


# We serialize dictionaries as json, and json doesn't know how to serialize AssetKeys. So we wrap the mapping
# to be able to serialize this dictionary with "non scalar" keys.
class CacheableSpecMappingSerializer(FieldSerializer):
    def pack(
        self,
        mapping: Mapping[str, "CacheableAssetSpec"],
        whitelist_map: WhitelistMap,
        descent_path: str,
    ) -> JsonSerializableValue:
        return pack_value(SerializableNonScalarKeyMapping(mapping), whitelist_map, descent_path)

    def unpack(
        self,
        unpacked_value: JsonSerializableValue,
        whitelist_map: WhitelistMap,
        context: UnpackContext,
    ) -> PackableValue:
        return unpack_value(unpacked_value, dict, whitelist_map, context)


@whitelist_for_serdes(
    field_serializers={
        "task_asset_specs": CacheableSpecMappingSerializer,
        "dag_asset_specs": CacheableSpecMappingSerializer,
    }
)
@record
class CacheableSpecs:
    task_asset_specs: Dict[AssetKey, "CacheableAssetSpec"]
    dag_asset_specs: Dict[AssetKey, "CacheableAssetSpec"]


@whitelist_for_serdes
@record
class CacheableAssetSpec:
    # A dumbed down version of AssetSpec that can be serialized easily to and from a dictionary.
    asset_key: AssetKey
    description: Optional[str]
    metadata: Mapping[str, Any]
    tags: Mapping[str, str]
    deps: Optional[Sequence["CacheableAssetDep"]]
    group_name: Optional[str]

    def to_asset_spec(self, additional_metadata: Mapping[str, Any]) -> AssetSpec:
        return AssetSpec(
            key=self.asset_key,
            description=self.description,
            metadata={**additional_metadata, **self.metadata},
            tags=self.tags,
            deps=[AssetDep(asset=dep.asset_key) for dep in self.deps] if self.deps else [],
            group_name=self.group_name,
        )

    @staticmethod
    def from_asset_spec(asset_spec: AssetSpec) -> "CacheableAssetSpec":
        return CacheableAssetSpec(
            asset_key=asset_spec.key,
            description=asset_spec.description,
            metadata=asset_spec.metadata,
            tags=asset_spec.tags,
            deps=[CacheableAssetDep.from_asset_dep(dep) for dep in asset_spec.deps],
            group_name=asset_spec.group_name,
        )


@whitelist_for_serdes
@record
class CacheableAssetDep:
    # A dumbed down version of AssetDep that can be serialized easily to and from a dictionary.
    asset_key: AssetKey

    @staticmethod
    def from_asset_dep(asset_dep: AssetDep) -> "CacheableAssetDep":
        return CacheableAssetDep(asset_key=asset_dep.asset_key)


DEFAULT_POLL_INTERVAL = 60


class AirflowCacheableAssetsDefinition(CacheableAssetsDefinition):
    """Builds cacheable assets definitions for Airflow DAGs."""

    def __init__(
        self,
        airflow_instance: AirflowInstance,
        poll_interval: int,
        defs: Optional[Definitions] = None,
        migration_state_override: Optional[AirflowMigrationState] = None,
    ):
        self.airflow_instance = airflow_instance
        self.poll_interval = poll_interval
        self.defs = defs
        self.migration_state_override = migration_state_override

    @property
    def unique_id(self) -> str:
        airflow_instance_name_hash = hashlib.md5(
            self.airflow_instance.normalized_name.encode()
        ).hexdigest()
        return f"airflow_assets_{airflow_instance_name_hash}"

    def compute_cacheable_data(self) -> Sequence[AssetsDefinitionCacheableData]:
        dag_infos = {dag.dag_id: dag for dag in self.airflow_instance.list_dags()}
        cacheable_task_data = construct_cacheable_assets_and_infer_dependencies(
            definitions=self.defs,
            migration_state=self.migration_state_override,
            airflow_instance=self.airflow_instance,
            dag_infos=dag_infos,
        )

        dag_specs_per_key: Dict[AssetKey, CacheableAssetSpec] = {}
        for dag in dag_infos.values():
            source_code = self.airflow_instance.get_dag_source_code(dag.metadata["file_token"])
            dag_specs_per_key[dag.dag_asset_key] = get_cached_spec_for_dag(
                airflow_instance=self.airflow_instance,
                task_asset_keys_in_dag=cacheable_task_data.all_asset_keys_per_dag_id.get(
                    dag.dag_id, set()
                ),
                downstreams_asset_dependency_graph=cacheable_task_data.downstreams_asset_dependency_graph,
                dag_info=dag,
                source_code=source_code,
            )
        return [
            AssetsDefinitionCacheableData(
                extra_metadata={
                    "cacheable_specs": serialize_value(
                        CacheableSpecs(
                            task_asset_specs=cacheable_task_data.cacheable_specs_per_asset_key,
                            dag_asset_specs=dag_specs_per_key,
                        )
                    ),
                },
            )
        ]

    def build_definitions(
        self, data: Sequence[AssetsDefinitionCacheableData]
    ) -> Sequence[AssetsDefinition]:
        check.invariant(len(data) == 1, "Expected exactly one cacheable data object.")
        metadata: Mapping[str, Any] = check.not_none(
            data[0].extra_metadata, "Expected cacheable data to have extra_metadata field set."
        )
        cacheable_specs = deserialize_value(metadata["cacheable_specs"], CacheableSpecs)

        new_assets_defs = []
        for dag_spec in cacheable_specs.dag_asset_specs.values():
            key = dag_spec.asset_key
            dag_id = dag_spec.metadata["Dag ID"]
            new_assets_defs.append(
                build_airflow_asset_from_specs(
                    specs=[dag_spec.to_asset_spec({})],
                    name=convert_to_valid_dagster_name(key.to_user_string()),
                    tags={DAG_ID_TAG: dag_id},
                )
            )
        return new_assets_defs + construct_assets_with_task_migration_info_applied(
            definitions=self.defs,
            cacheable_specs=cacheable_specs,
        )


def get_cached_spec_for_dag(
    airflow_instance: AirflowInstance,
    task_asset_keys_in_dag: Set[AssetKey],
    downstreams_asset_dependency_graph: Dict[AssetKey, Set[AssetKey]],
    dag_info: DagInfo,
    source_code: str,
) -> CacheableAssetSpec:
    leaf_asset_keys = get_leaf_assets_for_dag(
        asset_keys_in_dag=task_asset_keys_in_dag,
        downstreams_asset_dependency_graph=downstreams_asset_dependency_graph,
    )
    metadata = {
        "Dag Info (raw)": JsonMetadataValue(dag_info.metadata),
        "Dag ID": dag_info.dag_id,
        "Link to DAG": UrlMetadataValue(dag_info.url),
    }
    # Attempt to retrieve source code from the DAG.
    metadata["Source Code"] = MarkdownMetadataValue(
        f"""
```python
{source_code}
```
            """
    )

    return CacheableAssetSpec(
        asset_key=dag_info.dag_asset_key,
        description=f"A materialization corresponds to a successful run of airflow DAG {dag_info.dag_id}.",
        metadata=metadata,
        tags={"dagster/compute_kind": "airflow", DAG_ID_TAG: dag_info.dag_id},
        deps=[CacheableAssetDep(asset_key=key) for key in leaf_asset_keys],
        group_name=None,
    )


def build_airflow_asset_from_specs(
    specs: Sequence[AssetSpec], name: str, tags: Mapping[str, str]
) -> AssetsDefinition:
    @multi_asset(
        name=name,
        specs=specs,
        compute_kind="airflow",
        op_tags=tags,
    )
    def _asset(context):
        raise Exception(
            "Attempted to materialize a non-executable Airflow asset."
            " This may happen when the Dagster migration state does not match the Airflow migration state."
            " You may need to reload your Dagster code location."
        )

    return _asset


@record
class _CacheableData:
    cacheable_specs_per_asset_key: Dict[AssetKey, CacheableAssetSpec] = {}
    all_asset_keys_per_dag_id: Dict[str, Set[AssetKey]] = {}
    downstreams_asset_dependency_graph: Dict[AssetKey, Set[AssetKey]] = {}


def construct_cacheable_assets_and_infer_dependencies(
    definitions: Optional[Definitions],
    migration_state: Optional[AirflowMigrationState],
    airflow_instance: AirflowInstance,
    dag_infos: Dict[str, DagInfo],
) -> _CacheableData:
    downstreams_asset_dependency_graph: Dict[AssetKey, Set[AssetKey]] = defaultdict(set)
    cacheable_specs_per_asset_key: Dict[AssetKey, CacheableAssetSpec] = {}
    all_asset_keys_per_dag_id: Dict[str, Set[AssetKey]] = defaultdict(set)
    if not definitions or not definitions.assets:
        return _CacheableData()
    for asset in definitions.assets:
        asset = check.inst(  # noqa: PLW2901
            asset,
            (AssetsDefinition, AssetSpec),
            "Expected orchestrated defs to all be AssetsDefinitions or AssetSpecs.",
        )
        task_info = get_task_info_for_asset(airflow_instance, asset)
        if task_info is None:
            continue
        task_level_metadata = {
            "Task Info (raw)": JsonMetadataValue(task_info.metadata),
            # In this case,
            "Dag ID": task_info.dag_id,
            "Link to DAG": UrlMetadataValue(task_info.dag_url),
        }
        migration_state_for_task = _get_migration_state_for_task(
            migration_state_override=migration_state,
            task_info=task_info,
            dag_info=dag_infos[task_info.dag_id],
        )
        task_level_metadata[
            "Computed in Task ID" if migration_state_for_task is False else "Triggered by Task ID"
        ] = task_info.task_id
        specs = asset.specs if isinstance(asset, AssetsDefinition) else [asset]
        for spec in specs:
            spec_deps = []
            for dep in spec.deps:
                spec_deps.append(CacheableAssetDep.from_asset_dep(dep))
                downstreams_asset_dependency_graph[dep.asset_key].add(spec.key)
            cacheable_specs_per_asset_key[spec.key] = CacheableAssetSpec(
                asset_key=spec.key,
                description=spec.description,
                metadata=task_level_metadata,
                tags={
                    **spec.tags,
                    MIGRATED_TAG: str(migration_state_for_task),
                    DAG_ID_TAG: task_info.dag_id,
                    TASK_ID_TAG: task_info.task_id,
                },
                deps=spec_deps,
                group_name=spec.group_name,
            )
            all_asset_keys_per_dag_id[task_info.dag_id].add(spec.key)
    return _CacheableData(
        cacheable_specs_per_asset_key=cacheable_specs_per_asset_key,
        all_asset_keys_per_dag_id=all_asset_keys_per_dag_id,
        downstreams_asset_dependency_graph=downstreams_asset_dependency_graph,
    )


def construct_assets_with_task_migration_info_applied(
    definitions: Optional[Definitions],
    cacheable_specs: CacheableSpecs,
) -> List[AssetsDefinition]:
    if not definitions or not definitions.assets:
        return []

    new_assets_defs = []
    for asset in definitions.assets:
        asset = check.inst(  # noqa: PLW2901
            asset,
            (AssetSpec, AssetsDefinition),
            "Expected orchestrated defs to all be AssetsDefinitions or AssetSpecs.",
        )
        dag_id = get_dag_id_from_asset(asset)
        if dag_id is None:
            new_assets_defs.append(asset)
            continue
        overall_migration_status = None
        overall_task_id = None
        overall_dag_id = None
        new_specs = []
        specs = asset.specs if isinstance(asset, AssetsDefinition) else [asset]
        for spec in specs:
            check.invariant(
                cacheable_specs.task_asset_specs.get(spec.key) is not None,
                f"Could not find cacheable spec for asset key {spec.key.to_user_string()}",
            )
            # We allow arbitrary (non-serdes) metadata in asset specs, which makes them non-serializable.
            # This means we need to "combine" the metadata from the cacheable spec with the non-serializable
            # metadata fields from the original spec it was built from after deserializing.
            new_spec = cacheable_specs.task_asset_specs[spec.key].to_asset_spec(spec.metadata)
            check.invariant(
                MIGRATED_TAG in new_spec.tags,
                f"Could not find migrated status for asset key {spec.key.to_user_string()}",
            )
            check.invariant(
                overall_migration_status is None
                or overall_migration_status == new_spec.tags[MIGRATED_TAG],
                "Expected all assets in an AssetsDefinition (and therefore the same task) to have the same migration status.",
            )
            overall_migration_status = new_spec.tags[MIGRATED_TAG]
            check.invariant(
                DAG_ID_TAG in new_spec.tags,
                f"Could not find dag ID for asset key {spec.key.to_user_string()}",
            )
            dag_id = new_spec.tags[DAG_ID_TAG]
            check.invariant(
                overall_dag_id is None or overall_dag_id == dag_id,
                "Expected all assets in an AssetsDefinition to have the same dag ID.",
            )
            overall_dag_id = dag_id
            check.invariant(
                TASK_ID_TAG in new_spec.tags,
                f"Could not find task ID for asset key {spec.key.to_user_string()}",
            )
            task_id = new_spec.tags[TASK_ID_TAG]
            check.invariant(
                overall_task_id is None or overall_task_id == task_id,
                f"Expected all assets in an AssetsDefinition to have the same task ID. Found {overall_task_id} and {task_id}.",
            )
            overall_task_id = task_id
            new_specs.append(new_spec)

        # We don't coerce to a bool here since bool("False") is True.
        if overall_migration_status == "True":
            asset = check.inst(  # noqa: PLW2901
                asset,
                AssetsDefinition,
                f"For an asset to be migrated, it must be an AssetsDefinition. Found an AssetSpec for task ID {task_id}.",
            )
            check.invariant(
                asset.is_executable,
                f"For an asset to be marked as migrated, it must also be executable in dagster. Found unexecutable assets for task ID {task_id}.",
            )
            # This is suspect. Should we not be using a full AssetsDefinition copy here.
            # https://linear.app/dagster-labs/issue/FOU-372/investigate-suspect-code-in-construct-assets-with-task-migration-info
            new_assets_defs.append(
                AssetsDefinition(
                    keys_by_input_name=asset.keys_by_input_name,
                    keys_by_output_name=asset.keys_by_output_name,
                    node_def=asset.node_def,
                    partitions_def=asset.partitions_def,
                    specs=new_specs,
                )
            )
        else:
            new_assets_defs.append(
                build_airflow_asset_from_specs(
                    specs=new_specs,
                    name=convert_to_valid_dagster_name(f"{overall_dag_id}__{overall_task_id}"),
                    tags=asset.node_def.tags if isinstance(asset, AssetsDefinition) else {},
                )
            )
    return new_assets_defs


# We expect that every asset which is passed to this function has all relevant specs mapped to a task.
def get_task_info_for_asset(
    airflow_instance: AirflowInstance, asset: Union[AssetsDefinition, AssetSpec]
) -> Optional[TaskInfo]:
    task_id = get_task_id_from_asset(asset)
    dag_id = get_dag_id_from_asset(asset)
    if task_id is None or dag_id is None:
        return None
    return airflow_instance.get_task_info(dag_id, task_id)


def get_leaf_assets_for_dag(
    asset_keys_in_dag: Set[AssetKey],
    downstreams_asset_dependency_graph: Dict[AssetKey, Set[AssetKey]],
) -> List[AssetKey]:
    # An asset is a "leaf" for the dag if it has no transitive dependencies _within_ the dag. It may have
    # dependencies _outside_ the dag.
    leaf_assets = []
    cache = {}
    for asset_key in asset_keys_in_dag:
        if (
            get_transitive_dependencies_for_asset(
                asset_key, downstreams_asset_dependency_graph, cache
            ).intersection(asset_keys_in_dag)
            == set()
        ):
            leaf_assets.append(asset_key)
    return leaf_assets


def get_transitive_dependencies_for_asset(
    asset_key: AssetKey,
    downstreams_asset_dependency_graph: Dict[AssetKey, Set[AssetKey]],
    cache: Dict[AssetKey, Set[AssetKey]],
) -> Set[AssetKey]:
    if asset_key in cache:
        return cache[asset_key]
    transitive_deps = set()
    for dep in downstreams_asset_dependency_graph[asset_key]:
        transitive_deps.add(dep)
        transitive_deps.update(
            get_transitive_dependencies_for_asset(dep, downstreams_asset_dependency_graph, cache)
        )
    cache[asset_key] = transitive_deps
    return transitive_deps


def _get_migration_state_for_task(
    migration_state_override: Optional[AirflowMigrationState],
    task_info: TaskInfo,
    dag_info: DagInfo,
) -> bool:
    task_id = task_info.task_id
    dag_id = dag_info.dag_id
    if migration_state_override:
        return migration_state_override.get_migration_state_for_task(dag_id, task_id) or False
    else:
        return dag_info.migration_state.is_task_migrated(task_id) or False



from pathlib import Path

from setuptools import find_packages, setup

airflow_dep_list = [
    "apache-airflow>=2.0.0,<2.8",
    # Flask-session 0.6 is incompatible with certain airflow-provided test
    # utilities.
    "flask-session<0.6.0",
    "connexion<3.0.0",  # https://github.com/apache/airflow/issues/35234
    "pendulum>=2.0.0,<3.0.0",
]

setup(
    name="dagster-airlift",
    version="0.0.14",
    author="Dagster Labs",
    author_email="hello@dagsterlabs.com",
    license="Apache-2.0",
    description="Tooling to assist with migrating from Airflow to Dagster.",
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    url=(
        "https://github.com/dagster-io/dagster/tree/master/python_modules/libraries/"
        "dagster-airlift"
    ),
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(exclude=["dagster_airlift_tests*", "examples*"]),
    extras_require={
        "core": [
            "dagster",
        ],
        "in-airflow": airflow_dep_list,
        "mwaa": ["boto3"],
        "dbt": ["dagster-dbt"],
        "test": ["pytest", "dagster-dbt", "dbt-duckdb", "boto3", "dagster-webserver"],
    },
    zip_safe=False,
)



import os
import signal
import subprocess
import time
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Callable, Generator

import pytest
import requests
from dagster._core.test_utils import environ
from dagster._time import get_current_timestamp


####################################################################################################
# AIRFLOW SETUP FIXTURES
# Sets up the airflow environment for testing. Running at localhost:8080.
# Callsites are expected to provide implementations for dags_dir fixture.
####################################################################################################
def _airflow_is_ready():
    try:
        response = requests.get("http://localhost:8080")
        return response.status_code == 200
    except:
        return False


@pytest.fixture(name="airflow_home")
def default_airflow_home() -> Generator[str, None, None]:
    with TemporaryDirectory() as tmpdir:
        with environ({"AIRFLOW_HOME": tmpdir}):
            yield tmpdir


@pytest.fixture(name="setup")
def setup_fixture(airflow_home: Path, dags_dir: Path) -> Generator[Path, None, None]:
    assert os.environ["AIRFLOW_HOME"] == str(airflow_home), "AIRFLOW_HOME is not set correctly"
    temp_env = {
        **os.environ.copy(),
        # Provide link to local dagster installation.
        "DAGSTER_URL": "http://localhost:3333",
    }
    path_to_script = Path(__file__).parent.parent.parent / "scripts" / "airflow_setup.sh"
    subprocess.run(["chmod", "+x", path_to_script], check=True, env=temp_env)
    subprocess.run([path_to_script, dags_dir], check=True, env=temp_env)
    with environ(temp_env):
        yield airflow_home


@pytest.fixture(name="reserialize_dags")
def reserialize_fixture(airflow_instance: None) -> Callable[[], None]:
    """Forces airflow to reserialize dags, to ensure that the latest changes are picked up."""

    def _reserialize_dags() -> None:
        subprocess.check_output(["airflow", "dags", "reserialize"])

    return _reserialize_dags


@pytest.fixture(name="airflow_instance")
def airflow_instance_fixture(setup: None) -> Generator[subprocess.Popen, None, None]:
    process = subprocess.Popen(
        ["airflow", "standalone"],
        env=os.environ,  # since we have some temp vars in the env right now
        shell=False,
        preexec_fn=os.setsid,  # noqa  # fuck it we ball
    )
    # Give airflow a second to stand up
    time.sleep(5)
    initial_time = get_current_timestamp()

    airflow_ready = False
    while get_current_timestamp() - initial_time < 60:
        if _airflow_is_ready():
            airflow_ready = True
            break
        time.sleep(1)

    assert airflow_ready, "Airflow did not start within 30 seconds..."
    yield process
    # Kill process group, since process.kill and process.terminate do not work.
    os.killpg(process.pid, signal.SIGKILL)


####################################################################################################
# DAGSTER SETUP FIXTURES
# Sets up the dagster environment for testing. Running at localhost:3333.
# Callsites are expected to provide implementations for dagster_defs_path fixture.
####################################################################################################
def _dagster_is_ready() -> bool:
    try:
        response = requests.get("http://localhost:3333")
        return response.status_code == 200
    except:
        return False


@pytest.fixture(name="dagster_home")
def setup_dagster_home() -> Generator[str, None, None]:
    """Instantiate a temporary directory to serve as the DAGSTER_HOME."""
    with TemporaryDirectory() as tmpdir:
        with environ({"DAGSTER_HOME": tmpdir}):
            yield tmpdir


@pytest.fixture(name="dagster_dev")
def setup_dagster(dagster_home: str, dagster_defs_path: str) -> Generator[Any, None, None]:
    """Stands up a dagster instance using the dagster dev CLI. dagster_defs_path must be provided
    by a fixture included in the callsite.
    """
    process = subprocess.Popen(
        ["dagster", "dev", "-f", dagster_defs_path, "-p", "3333"],
        env=os.environ.copy(),
        shell=False,
        preexec_fn=os.setsid,  # noqa
    )
    # Give dagster a second to stand up
    time.sleep(5)

    dagster_ready = False
    initial_time = get_current_timestamp()
    while get_current_timestamp() - initial_time < 60:
        if _dagster_is_ready():
            dagster_ready = True
            break
        time.sleep(1)

    assert dagster_ready, "Dagster did not start within 30 seconds..."
    yield process
    os.killpg(process.pid, signal.SIGKILL)



from pathlib import Path
from typing import Any, Dict, NamedTuple, Optional, Sequence

import yaml


class TaskMigrationState(NamedTuple):
    task_id: str
    migrated: bool

    @staticmethod
    def from_dict(task_dict: Dict[str, Any]) -> "TaskMigrationState":
        if set(task_dict.keys()) != {"id", "migrated"}:
            raise Exception(
                f"Expected 'migrated' and 'id' keys in the task dictionary. Found keys: {task_dict.keys()}"
            )
        if task_dict["migrated"] not in [True, False]:
            raise Exception("Expected 'migrated' key to be a boolean")
        return TaskMigrationState(task_id=task_dict["id"], migrated=task_dict["migrated"])


class DagMigrationState(NamedTuple):
    tasks: Dict[str, TaskMigrationState]

    @staticmethod
    def from_dict(dag_dict: Dict[str, Sequence[Dict[str, Any]]]) -> "DagMigrationState":
        if "tasks" not in dag_dict:
            raise Exception(
                f"Expected a 'tasks' key in the dag dictionary. Instead; got: {dag_dict}"
            )
        task_list = dag_dict["tasks"]
        task_migration_states = {}
        for task_dict in task_list:
            task_state = TaskMigrationState.from_dict(task_dict)
            task_migration_states[task_state.task_id] = task_state
        return DagMigrationState(tasks=task_migration_states)

    def is_task_migrated(self, task_id: str) -> bool:
        if task_id not in self.tasks:
            return False
        return self.tasks[task_id].migrated


class AirflowMigrationState(NamedTuple):
    dags: Dict[str, DagMigrationState]

    def get_migration_state_for_task(self, dag_id: str, task_id: str) -> Optional[bool]:
        if dag_id not in self.dags:
            return None
        if task_id not in self.dags[dag_id].tasks:
            return None
        return self.dags[dag_id].tasks[task_id].migrated

    def dag_has_migration_state(self, dag_id: str) -> bool:
        return self.get_migration_dict_for_dag(dag_id) is not None

    def get_migration_dict_for_dag(
        self, dag_id: str
    ) -> Optional[Dict[str, Sequence[Dict[str, Any]]]]:
        if dag_id not in self.dags:
            return None
        return {
            "tasks": [
                {"migrated": task_state.migrated, "id": task_id}
                for task_id, task_state in self.dags[dag_id].tasks.items()
            ]
        }


class MigrationStateParsingError(Exception):
    pass


def load_migration_state_from_yaml(migration_yaml_path: Path) -> AirflowMigrationState:
    # Expect migration_yaml_path to be a directory, where each file represents a dag, and each
    # file in the subdir represents a task. The dictionary for each task should contain two keys;
    # id: the task id, and migrated: a boolean indicating whether the task has been migrated.
    dag_migration_states = {}
    try:
        for dag_file in migration_yaml_path.iterdir():
            # Check that the file is a yaml file or yml file
            if dag_file.suffix not in [".yaml", ".yml"]:
                continue
            dag_id = dag_file.stem
            yaml_dict = yaml.safe_load(dag_file.read_text())
            if not isinstance(yaml_dict, dict):
                raise Exception("Expected a dictionary")
            dag_migration_states[dag_id] = DagMigrationState.from_dict(yaml_dict)
    except Exception as e:
        raise MigrationStateParsingError("Error parsing migration yaml") from e
    return AirflowMigrationState(dags=dag_migration_states)
